// -*- C++ -*-
/*!
 * @file  RTC_GameController.cpp
 * @brief PS4&XBOX360?
 * @date $Date$
 *
 * $Id$
 */

#include "RTC_GameController.h"
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <cstdio>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/joystick.h>

//#define debug
int gp_oc=-1;
char SNC[50]="Sony Navigation Controller";
char WireC[50]="Wireless Controller";
char BOX[50]="Xbox 360 Wireless Receiver";
char DEV_NAME[512]="!!";
double pre_AXES[50];
int AXES_NUM=0;
int BUTTON_NUM=0;
int  Device_MODE=0;
struct GC_state GC_state;
struct js_event GC;
// Module specification
// <rtc-template block="module_spec">
static const char* rtc_gamecontroller_spec[] =
  {
    "implementation_id", "RTC_GameController",
    "type_name",         "RTC_GameController",
    "description",       "PS4&XBOX360",
    "version",           "1.0.0",
    "vendor",            "kohno,AIZUK",
    "category",          "Controller",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.K_Stick", "1.0",
    "conf.default.K_Va", "1.0",
    "conf.default.K_Vy", "1.0",
    "conf.default.K_Vx", "1.0",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RTC_GameController::RTC_GameController(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_ButtonOut("Button", m_Button),
    m_L_Stick_XOut("L_Stick_X", m_L_Stick_X),
    m_L_Stick_YOut("L_Stick_Y", m_L_Stick_Y),
    m_R_Stick_XOut("R_Stick_X", m_R_Stick_X),
    m_R_Stick_YOut("R_Stick_Y", m_R_Stick_Y),
    m_L2_AnalogOut("L2_Analog", m_L2_Analog),
    m_R2_AnalogOut("R2_Analog", m_R2_Analog),
    m_VelocityOut("Velocity2D", m_Velocity),
    m_AnalogOut("Analog",m_Analog),
    m_Controller_TypeOut("Pad_Type",m_Controller_Type)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
RTC_GameController::~RTC_GameController()
{
}



RTC::ReturnCode_t RTC_GameController::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  
  // Set OutPort buffer
  addOutPort("Controler_Type", m_Controller_TypeOut);
  addOutPort("Button", m_ButtonOut);
  addOutPort("Analog", m_AnalogOut);
  //addOutPort("L_Stick_X", m_L_Stick_XOut);
  //addOutPort("L_Stick_Y", m_L_Stick_YOut);
  //addOutPort("R_Stick_X", m_R_Stick_XOut);
  //addOutPort("R_Stick_Y", m_R_Stick_YOut);
  //addOutPort("L2_Analog", m_L2_AnalogOut);
  //addOutPort("R2_Analog", m_R2_AnalogOut);
  //addOutPort("Velocity2D", m_VelocityOut);
  bindParameter("K_Vx", m_K_Vx, "1.0");
  bindParameter("K_Vy", m_K_Vy, "1.0");
  bindParameter("K_Va", m_K_Va, "1.0");
  bindParameter("K_Stick", m_K_Stick, "1.0");
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // </rtc-template>
    m_Analog.data.length(11);
  
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RTC_GameController::onFinalize()
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RTC_GameController::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RTC_GameController::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RTC_GameController::onActivated(RTC::UniqueId ec_id)
{ 
  Game_Controller_Type_Find("/dev/input/js0");
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_GameController::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_GameController::onExecute(RTC::UniqueId ec_id)
{
	double rv=0,lv=0;
	double vx=0,vy=0,va=0;
	double axes_ud=0,axes_lr=0;

	m_Button.data=Game_Controller_Button_Read();
	 // Game_Controller_AXES_Read();
	#ifdef debug
		std::cerr<<"\x1b[1m" <<"\x1b[36m" << GC_state.AXES[1]<<" "<<GC_state.AXES[0]<<"\x1b[39m"<<"\x1b[0m"<< std::endl ;
		std::cerr<<"\x1b[1m" <<"\x1b[36m" << GC_state.AXES[3]<<" "<<GC_state.AXES[2]<<"\x1b[39m"<<"\x1b[0m"<< std::endl ;
		std::cerr<<"\x1b[1m" <<"\x1b[36m" << GC_state.AXES[5]<<" "<<GC_state.AXES[4]<<"\x1b[39m"<<"\x1b[0m"<< std::endl ;
		std::cerr<<"\x1b[1m" <<"\x1b[36m" << GC_state.AXES[7]<<" "<<GC_state.AXES[6]<<"\x1b[39m"<<"\x1b[0m"<< std::endl ;

	std::cerr<<"\x1b[1m" <<"\x1b[34m" << m_Button.data<<"\x1b[39m"<<"\x1b[0m" ;
	#endif
	printf("  %x\n", m_Button.data);

	if(Device_MODE==0){	//PS4
		m_Controller_Type.data = CORBA::string_dup(WireC);
		//std::cout <<m_Controller_Type.data<<std::endl;
	
		m_L_Stick_X.data=GC_state.AXES[1];
		m_L_Stick_Y.data=GC_state.AXES[0];
		m_R_Stick_X.data=GC_state.AXES[5];
		m_R_Stick_Y.data=GC_state.AXES[2];
		m_L2_Analog.data=GC_state.AXES[3];
		m_R2_Analog.data=GC_state.AXES[4];
		GC_state.HAT[0]=GC_state.AXES[7];
		GC_state.HAT[1]=GC_state.AXES[6];

		#ifdef debug
		std::cerr<<"\x1b[1m" <<"\x1b[36m" <<GC_state.HAT[0]<<" "
		<<GC_state.HAT[1]<<"\x1b[39m"<<"\x1b[0m"<< std::endl ;
		std::cerr<<"\x1b[1m" <<"\x1b[36m" <<m_L_Stick_X.data<<" "
		<<m_L_Stick_Y.data<<"\x1b[39m"<<"\x1b[0m"<< std::endl ;
		std::cerr<<"\x1b[1m" <<"\x1b[36m" <<m_R_Stick_X.data<<" "
		<<m_R_Stick_Y.data<<"\x1b[39m"<<"\x1b[0m"<< std::endl ;
		#endif
		//Stick
		axes_ud=-100*GC_state.AXES[1];//↑:-1 ↓:1
		axes_lr=100*GC_state.AXES[0];//←:-1 →:1
		printf(" \n");
		#ifdef debug
			std::cerr<<"\x1b[1m" <<"\x1b[36m" << GC_state.AXES[1]<<" "
			<< axes_lr <<"\x1b[39m"<<"\x1b[0m"<< std::endl ;
		#endif

		rv=axes_ud+axes_lr;
		lv=axes_ud-axes_lr;
		if(rv>100)rv=100;
		if(rv<-100)rv=-100;
		if(lv>100)lv=100;
		if(lv<-100)lv=-100;
		va=(rv-lv)/2;

		#ifdef debug
			printf(" \n");
			printf(" \x1b[1m \x1b[36m   [ rv ,  lv ]   =  [ %f  ,  %f ]  \x1b[39m   \x1b[0m  \n",rv,lv);
		#endif

		axes_ud=-100*GC_state.AXES[1];//↑:-1 ↓:1
		axes_lr=-100*GC_state.AXES[0];//←:-1 →:1
		vx=axes_ud;
		vy=axes_lr;

		#ifdef debug
			printf(" \n");
			printf(" \x1b[1m \x1b[32m   [ Vx ,  Vy , Va]   =  [ %f  ,  %f , %f ]  \x1b[39m   \x1b[0m  \n",vx,vy,va );
		#endif
		//std::cerr<<"\x1b[1m" <<"\x1b[36m" << rv <<" "<< lv <<"\x1b[39m"<<"\x1b[0m"<< std::endl ;
		
		
		m_L_Stick_X.data=GC_state.AXES[1];
		m_L_Stick_Y.data=GC_state.AXES[0];
		m_R_Stick_X.data=GC_state.AXES[5];
		m_R_Stick_Y.data=GC_state.AXES[2];
		m_L2_Analog.data=GC_state.AXES[3];
		m_R2_Analog.data=GC_state.AXES[4];
		
		m_Analog.data[0]=GC_state.HAT[0];
		m_Analog.data[1]=GC_state.HAT[1];
		m_Analog.data[2]=GC_state.AXES[1];
		m_Analog.data[3]=GC_state.AXES[0];
		m_Analog.data[4]=GC_state.AXES[5];
		m_Analog.data[5]=GC_state.AXES[2];
		m_Analog.data[6]=GC_state.AXES[3];
		m_Analog.data[7]=GC_state.AXES[4];
		m_Controller_TypeOut.write();
		
	}
	if(Device_MODE==1){	//Navigation
	}
	if(Device_MODE==2){	//XBOX360
	}

	//m_Velocity.data.vx=vx;
	//m_Velocity.data.vy=vy;
	//m_Velocity.data.va=va;
	/*
	m_L_Stick_XOut.write();
	m_L_Stick_YOut.write();
	m_R_Stick_XOut.write();Game_Controller_Button_Read
	m_R_Stick_YOut.write();
	m_L2_AnalogOut.write();
	m_R2_AnalogOut.write();
	m_VelocityOut.write();*/
	m_ButtonOut.write();
	m_AnalogOut.write();

	return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RTC_GameController::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RTC_GameController::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RTC_GameController::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RTC_GameController::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RTC_GameController::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

int Game_Controller_Type_Find(const char* GC_dev){

	 if( ( gp_oc=open(GC_dev,O_RDONLY) ) < 0){
		std::cerr <<"\x1b[31m" <<"DEVICE OPEN ERROR"  << std::endl;
		std::cerr << "Prease Chek Connection"  << "\x1b[39m"<<std::endl;
		exit(0);
		//return 0;
	}

	std::cout << GC_dev  << std::endl;/*
	ioctl(fd, JSIOCGVERSION, &version);*/
	ioctl(gp_oc, JSIOCGAXES, &AXES_NUM);
	ioctl(gp_oc, JSIOCGBUTTONS, &BUTTON_NUM);
	ioctl(gp_oc, JSIOCGNAME(256), DEV_NAME);
	fcntl(gp_oc, F_SETFL, O_NONBLOCK);
	//std::cout <<"V :"<< JSIOCGVERSION  <<" A :"<<JSIOCGAXES  <<" B :"<<JSIOCGBUTTONS  <<" N :"<<dev_name<<"   "<< std::endl;

	if( strcmp(DEV_NAME,WireC)==0 ){
		std::cout <<"Wireless Controller"<<std::endl;
		Device_MODE=0;
	}else if( strcmp(DEV_NAME,SNC)==0 ){
		std::cout <<"Sony Navigation Controler"<<std::endl;
		Device_MODE=1;
	}else if( strcmp(DEV_NAME,BOX)==0 ){
		std::cout <<"Xbox 360 Wireless Receiver"<<std::endl;
		Device_MODE=2;
	}

return 1;
}

int Game_Controller_AXES_Read(){
return 1;
}

unsigned long Game_Controller_Button_Read(){
	int i=0;
	unsigned long but=0;
 //JPD js;
//    for(i=0;i<BUTTON_NUM;i++)
//	joy_button[i]=0;
//	std::cout <<"AXES:"<<AXES_NUM<<"BUTTON:"<<BUTTON_NUM<<std::endl;
   // read(gp_oc, &GC, sizeof(GC));
      unsigned int r = 0;
      while (r < sizeof(struct js_event)){
		r +=read(gp_oc, &GC, sizeof(GC));
		switch (GC.type & ~JS_EVENT_INIT)
		{
		case JS_EVENT_BUTTON:
		if((int)GC.number>=BUTTON_NUM)  {
		//     cerr<<"err:"<<(int)js.number<<endl;
		//       continue;
		}
		GC_state.BUTTON[(int)GC.number]= GC.value;
		break;

		case JS_EVENT_AXIS:
		if((int)GC.number>=AXES_NUM)  {
		//  cerr<<"err:"<<(int)js.number<<endl;
		//       continue;
		}
		GC_state.Pre_AXES[(int)GC.number]= GC.value;

		break;
		}
	}

   for(i=0;i<AXES_NUM;++i){
	if(Device_MODE==0){
	    GC_state.AXES[i]=(float)GC_state.Pre_AXES[i]/32767;
#ifdef debug
	    printf("%f ", GC_state.AXES[i]);
#endif
	    }
	if(Device_MODE==1){
	    GC_state.AXES[i]=(float)GC_state.Pre_AXES[i]/32767;
#ifdef debug
	    printf("%f ", GC_state.AXES[i]);
#endif
	    }
	if(Device_MODE==2){
	    GC_state.AXES[i]=(float)GC_state.Pre_AXES[i]/32767;
#ifdef debug
	    printf("%f ",GC_state.AXES[i]);
#endif
	    }
    }



#ifdef debug
std::cout <<"BUTTON";
#endif
	for(i=0;i<BUTTON_NUM;++i){
		if(Device_MODE==0){
		#ifdef debug
		   printf("%d ",(int)GC_state.BUTTON[i]);
		#endif
		if((int)GC_state.BUTTON[i] == 1)
		   but |= (1<<i);
		    }
		if(Device_MODE==1){
		#ifdef debug
		    printf("%d ", (int)GC_state.BUTTON[i]);
		#endif
		if((int)GC_state.BUTTON[i] == 1)
		   but |= (1<<i);
	
		    }
		if(Device_MODE==2){
		#ifdef debug
		    printf("%d ", (int)GC_state.BUTTON[i]);
		#endif
		if((int)GC_state.BUTTON[i] == 1)
		   but |= (1<<i);

		    }
	}
 
#ifdef debug
    printf("\n");
#endif
    
    return but;
}

int Game_Controller_Close(const char* device){
return 1;
}



extern "C"
{
 
  void RTC_GameControllerInit(RTC::Manager* manager)
  {
    coil::Properties profile(rtc_gamecontroller_spec);
    manager->registerFactory(profile,
                             RTC::Create<RTC_GameController>,
                             RTC::Delete<RTC_GameController>);
  }
  
};


