// -*- C++ -*-
/*!
 * @file  RTC_GameController.cpp
 * @brief PS4&XBOX360?
 * @date $Date$
 *
 * $Id$
 */

#include "RTC_GameController.h"
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <cstdio>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/joystick.h>

//#define debug

//...................................................//
//接続用
//...................................................//
int gp_oc=-1;
char DEV_NAME[512]="!!";
char* portname;

//...................................................//
//デバイス名判定用
//...................................................//
char SNC[50]="Sony Navigation Controller";
char WireC[50]="Wireless Controller";
char BOX[50]="Xbox 360 Wireless Receiver";

//...................................................//
//ボタン・アナログ用
//...................................................//
double pre_AXES[50];
int AXES_NUM=0;
int BUTTON_NUM=0;
int  Device_MODE=0;
int  Device_CONECT=1;
struct GC_state GC_state;
struct js_event GC;
// Module specification
// <rtc-template block="module_spec">
static const char* rtc_gamecontroller_spec[] =
  {
    "implementation_id", "RTC_GameController",
    "type_name",         "RTC_GameController",
    "description",       "PS4コントローラの情報取得",
    "version",           "1.0.1",
    "vendor",            "kohno,AIZUK",
    "category",          "Controller",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.Port", "/dev/input/js0",
    "conf.default.Stick_Gain", "1.0",
    // Widget
    "conf.__widget__.Port", "text",
    "conf.__widget__.Stick_Gain", "text",
    // Constraints
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RTC_GameController::RTC_GameController(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_AnalogOut("Analog", m_Analog),
    m_ButtonOut("Button", m_Button),
    m_Controller_TypeOut("Controller_Type", m_Controller_Type)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
RTC_GameController::~RTC_GameController()
{
}



RTC::ReturnCode_t RTC_GameController::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  
  // Set OutPort buffer
  addOutPort("Controller_Type", m_Controller_TypeOut);
  addOutPort("Button", m_ButtonOut);
  addOutPort("Analog", m_AnalogOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("Port", m_Port, "/dev/input/js0");
  bindParameter("Stick_Gain", m_Stick_Gain, "1.0");
  // </rtc-template>
    m_Analog.data.length(11);
  
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RTC_GameController::onFinalize()
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RTC_GameController::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RTC_GameController::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RTC_GameController::onActivated(RTC::UniqueId ec_id)
{ 
  int p_len;
  int Open =1;
////////////////////////
//GamePAD PORT名取得・接続
///////////////////////

//接続できるまでLOOP
  while(Open){
	  //PORT名長さ取得
	  p_len=m_Port.length();
	  //PORT名格納先初期化
	  portname=new char[p_len+1];
	  //PORT名格納
	  memcpy(portname, m_Port.c_str(), p_len+1);
	  //PORT接続 失敗：１ 成功：０ while文抜ける
	  Open = Game_Controller_Type_Find(portname);
  }
  
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_GameController::onDeactivated(RTC::UniqueId ec_id)
{
  close(gp_oc);
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_GameController::onExecute(RTC::UniqueId ec_id)
{
	
	//GamePadボタン・アナログ取得
	m_Button.data=Game_Controller_Button_AXES_Read();
	
	#ifdef debug
		printf("Button OUT\t%.4x\n\r",m_Button.data);	
	#endif
	

	if(Device_MODE==0){	//PS4
		//GamePadデバイス名
	        m_Controller_Type.data = CORBA::string_dup(WireC);
		
		//GamePad HATキー格納  ものによってはアナログでない
		GC_state.HAT[0]=GC_state.AXES[7];
		GC_state.HAT[1]=GC_state.AXES[6];
	
		//GamePad ANALOGデータ格納
		// [2] 〜[5] (スティック)のみゲイン調整により出力範囲変更可能　m_Stick_Gain :（デフォルト = 1.0）
		//　[ -m_Stick_Gain  〜　m_Stick_Gain　]
		m_Analog.data[0]=GC_state.HAT[0];			//HAT Up Down
		m_Analog.data[1]=GC_state.HAT[1];			//HAT Lefft Right
		m_Analog.data[2]=GC_state.AXES[1] * m_Stick_Gain;	//Left Stick Up Down
		m_Analog.data[3]=GC_state.AXES[0] * m_Stick_Gain;	//Left Stick Lefft Right
		m_Analog.data[4]=GC_state.AXES[5] * m_Stick_Gain;	//Right Stick Up Down
		m_Analog.data[5]=GC_state.AXES[2] * m_Stick_Gain;	//Right Stick Lefft Right
		m_Analog.data[6]=GC_state.AXES[3] ;	//L2 Button
		m_Analog.data[7]=GC_state.AXES[4] ;	//R2 Button
		
	
	#ifdef debug
		std::cerr<<"ANALOG_OUT[0] [1]\t"<<
		m_Analog.data[0]<<" "<<m_Analog.data[1]<< std::endl ;
		std::cerr<<"ANALOG_OUT[2] [3]\t"<<
		m_Analog.data[2]<<" "<<m_Analog.data[3]<< std::endl ;
		std::cerr<<"ANALOG_OUT[4] [5]\t"<<
		m_Analog.data[4]<<" "<<m_Analog.data[5]<< std::endl ;
		std::cerr<<"ANALOG_OUT[6] [7]\t"<<
		m_Analog.data[6]<<" "<<m_Analog.data[7]<< std::endl ;
	#endif
		
	}
	if(Device_MODE==1){	//Navigation
	}
	if(Device_MODE==2){	//XBOX360
	}

	//GamePadデバイス名 PORT出力
	m_Controller_TypeOut.write();
	//GamePadボタン状態 PORT出力	
	m_ButtonOut.write();
	//GamePadアナログ状態 PORT出力
	m_AnalogOut.write();

	return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RTC_GameController::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RTC_GameController::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RTC_GameController::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RTC_GameController::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RTC_GameController::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

//
//GamePadへの接続とPad情報取得
//入力
//const char* GC_dev : 接続先PORT
//
int Game_Controller_Type_Find(const char* GC_dev){
//GamePadへの接続
	 if( ( gp_oc=open(GC_dev,O_RDONLY) ) < 0){
	 	//GamePadへの接続失敗
		std::cerr <<"\x1b[31m" <<"DEVICE OPEN ERROR"  << std::endl;
		std::cerr << "Prease Chek Connection"  << "\x1b[39m"<<std::endl;
		Device_CONECT=1;
		//exit(0);
		//return 0;
	}else{
		//GamePadへの接続成功
		Device_CONECT=0;
	}
	//接続先PORT表示
	std::cout << GC_dev  << std::endl;
	/*
	ioctl(fd, JSIOCGVERSION, &version);
	*/
	//PAD ANALOG数取得
	ioctl(gp_oc, JSIOCGAXES, &AXES_NUM);
	//PAD ボタン数取得
	ioctl(gp_oc, JSIOCGBUTTONS, &BUTTON_NUM);
	//PAD デバイス名取得
	ioctl(gp_oc, JSIOCGNAME(256), DEV_NAME);
	fcntl(gp_oc, F_SETFL, O_NONBLOCK);

	if( strcmp(DEV_NAME,WireC)==0 ){
		std::cout <<"Wireless Controller"<<std::endl;
		Device_MODE=0;
	}else if( strcmp(DEV_NAME,SNC)==0 ){
		std::cout <<"Sony Navigation Controler"<<std::endl;
		Device_MODE=1;
	}else if( strcmp(DEV_NAME,BOX)==0 ){
		std::cout <<"Xbox 360 Wireless Receiver"<<std::endl;
		Device_MODE=2;
	}

return Device_CONECT;
}

unsigned long Game_Controller_Button_AXES_Read(){
	int i=0;
	unsigned long but=0;
      unsigned int r = 0;
      
      while (r < sizeof(struct js_event)){
		r +=read(gp_oc, &GC, sizeof(GC));
		switch (GC.type & ~JS_EVENT_INIT)
		{
			case JS_EVENT_BUTTON://ボタンが押された
			GC_state.BUTTON[(int)GC.number]= GC.value;
			break;

			case JS_EVENT_AXIS://アナログ値が変化
			GC_state.Pre_AXES[(int)GC.number]= GC.value;

			break;
		}
	}

   for(i=0;i<AXES_NUM;++i){
	if(Device_MODE==0){
	    //アナログ値を ( −1.0 〜 1.0 ）の範囲に PS4
	    GC_state.AXES[i]=(float)GC_state.Pre_AXES[i]/32767;
#ifdef debug
//	    printf("PS4_AXIS[%d] %f \n\r", i,GC_state.AXES[i]);
#endif
	    }
	if(Device_MODE==1){
	    //アナログ値を ( −1.0 〜 1.0 ）の範囲に  要変更
	    GC_state.AXES[i]=(float)GC_state.Pre_AXES[i]/32767;
#ifdef debug
//	    printf("%f ", GC_state.AXES[i]);
#endif
	    }
	if(Device_MODE==2){
	    //アナログ値を ( −1.0 〜 1.0 ）の範囲に XBOX360 要変更
	    GC_state.AXES[i]=(float)GC_state.Pre_AXES[i]/32767;
#ifdef debug
//	    printf("XBOX360_AXIS[%d] %f ", i ,GC_state.AXES[i]);
#endif
	    }
    }



#ifdef debug
	std::cout<<"\x1b[1m\x1b[34m"<<"BUTTON ";
#endif

//何番のボタンが押されたか
//デバイス毎に違う 要 極力同一にする処理
	for(i=0;i<BUTTON_NUM;++i){
		if(Device_MODE==0){
		if((int)GC_state.BUTTON[i] == 1)
		   but |= (1<<i);
		    }
		if(Device_MODE==1){
		if((int)GC_state.BUTTON[i] == 1)
		   but |= (1<<i);
	
		    }
		if(Device_MODE==2){
		if((int)GC_state.BUTTON[i] == 1)
		   but |= (1<<i);

		    }
	}
 
#ifdef debug
	for(i=0;i<BUTTON_NUM;++i)
        printf("%d ",(int)GC_state.BUTTON[BUTTON_NUM-1-i]);
	printf("\x1b[39m\x1b[0m\n");
#endif
    
    return but;
}


extern "C"
{
 
  void RTC_GameControllerInit(RTC::Manager* manager)
  {
    coil::Properties profile(rtc_gamecontroller_spec);
    manager->registerFactory(profile,
                             RTC::Create<RTC_GameController>,
                             RTC::Delete<RTC_GameController>);
  }
  
};


